#include <nall/string.hpp>
#include <nall/vector.hpp>
#include <nall/windows/detour.hpp>
using namespace nall;

struct HeaderMagic {
  struct File {
    HANDLE handle;
    string name;
    bool header;
  };
  linear_vector<File> file;
};

HeaderMagic headerMagic;

auto oIsDebuggerPresent = IsDebuggerPresent;
auto oCreateFileA = CreateFileA;
auto oCloseHandle = CloseHandle;
auto oFindFirstFileA = FindFirstFileA;
auto oFindNextFileA = FindNextFileA;
auto oGetFileSize = GetFileSize;
auto oSetFilePointer = SetFilePointer;
auto oReadFile = ReadFile;
auto oWriteFile = WriteFile;

BOOL WINAPI pIsDebuggerPresent() {
  return false;
//return oIsDebuggerPresent();
}

HANDLE WINAPI pCreateFileA(LPCSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile) {
  HANDLE result = oCreateFileA(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile);
  if(lpFileName && result) {
    if(striend(lpFileName, ".sfc") || striend(lpFileName, ".smc")) {
      bool found = false;
      foreach(file, headerMagic.file) {
        if(file.name == lpFileName) { found = true; break; }
      }
      if(found == false) {
        HeaderMagic::File file;
        file.handle = result;
        file.name = lpFileName;
        file.header = (oGetFileSize(result, NULL) & 0x7fff) == 512;
        headerMagic.file.append(file);
      }
    }
  }
  return result;
}

BOOL WINAPI pCloseHandle(HANDLE hObject) {
  signed index = -1;
  foreach(file, headerMagic.file, n) {
    if(file.handle == hObject) { index = n; break; }
  }
  if(index >= 0) headerMagic.file.remove(index);
  return oCloseHandle(hObject);
}

HANDLE WINAPI pFindFirstFileA(LPCSTR lpFileName, LPWIN32_FIND_DATAA lpFindFileData) {
  HANDLE result = oFindFirstFileA(lpFileName, lpFindFileData);
  if(lpFindFileData) {
    if(striend(lpFindFileData->cFileName, ".sfc") || striend(lpFindFileData->cFileName, ".smc")) {
      if((lpFindFileData->nFileSizeLow & 0x7fff) == 0) lpFindFileData->nFileSizeLow -= 512;
    }
  }
  return result;
}

BOOL WINAPI pFindNextFileA(HANDLE hFindFile, LPWIN32_FIND_DATAA lpFindFileData) {
  BOOL result = oFindNextFileA(hFindFile, lpFindFileData);
  if(lpFindFileData) {
    if(striend(lpFindFileData->cFileName, ".sfc") || striend(lpFindFileData->cFileName, ".smc")) {
      if((lpFindFileData->nFileSizeLow & 0x7fff) == 0) lpFindFileData->nFileSizeLow -= 512;
    }
  }
  return result;
}

DWORD WINAPI pGetFileSize(HANDLE hFile, LPDWORD lpFileSizeHigh) {
  DWORD result = oGetFileSize(hFile, lpFileSizeHigh);
  foreach(file, headerMagic.file) {
    if(file.handle == hFile && file.header == false) return result + 512;
  }
  return result;
}

DWORD WINAPI pSetFilePointer(HANDLE hFile, LONG lDistanceToMove, PLONG lpDistanceToMoveHigh, DWORD dwMoveMethod) {
  return oSetFilePointer(hFile, lDistanceToMove, lpDistanceToMoveHigh, dwMoveMethod);
}

BOOL WINAPI pReadFile(HANDLE hFile, LPVOID lpBuffer, DWORD nNumberOfBytesToRead, LPDWORD lpNumberOfBytesRead, LPOVERLAPPED lpOverlapped) {
  bool displace = false;
  foreach(file, headerMagic.file) {
    if(file.handle == hFile) {
      if(file.header == false) displace = true;
      break;
    }
  }

  if(displace) {
    unsigned offset = oSetFilePointer(hFile, 0, 0, FILE_CURRENT);
    if(offset < 512) {
      unsigned length = min(512 - offset, nNumberOfBytesToRead);
      memset(lpBuffer, 0x00, length);
      lpBuffer = (LPVOID)((unsigned)lpBuffer + length);
      nNumberOfBytesToRead -= length;
      offset = 512;
      if(nNumberOfBytesToRead == 0) return TRUE;
    }
    oSetFilePointer(hFile, offset - 512, 0, FILE_BEGIN);
  }

  BOOL result = oReadFile(hFile, lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, lpOverlapped);

  if(displace) {
    unsigned offset = oSetFilePointer(hFile, 0, 0, FILE_CURRENT);
    oSetFilePointer(hFile, offset + 512, 0, FILE_BEGIN);
  }

  return result;
}

BOOL WINAPI pWriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped) {
  bool displace = false;
  foreach(file, headerMagic.file) {
    if(file.handle == hFile) {
      if(file.header == false) displace = true;
      break;
    }
  }

  if(displace) {
    unsigned offset = oSetFilePointer(hFile, 0, 0, FILE_CURRENT);
    if(offset < 512) {
      unsigned length = min(512 - offset, nNumberOfBytesToWrite);
      lpBuffer = (LPCVOID)((unsigned)lpBuffer + length);
      nNumberOfBytesToWrite -= length;
      offset = 512;
      if(nNumberOfBytesToWrite == 0) return TRUE;
    }
    oSetFilePointer(hFile, offset - 512, 0, FILE_BEGIN);
  }

  BOOL result = oWriteFile(hFile, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped);

  if(displace) {
    unsigned offset = oSetFilePointer(hFile, 0, 0, FILE_CURRENT);
    oSetFilePointer(hFile, offset + 512, 0, FILE_BEGIN);
  }

  return result;
}

extern "C" dllexport BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID reserved) {
  if(dwReason == DLL_PROCESS_ATTACH) {
    detour::insert("kernel32", "IsDebuggerPresent", (void*&)oIsDebuggerPresent, (void*)pIsDebuggerPresent);
    detour::insert("kernel32", "CreateFileA", (void*&)oCreateFileA, (void*)pCreateFileA);
    detour::insert("kernel32", "CloseHandle", (void*&)oCloseHandle, (void*)pCloseHandle);
    detour::insert("kernel32", "FindFirstFileA", (void*&)oFindFirstFileA, (void*)pFindFirstFileA);
    detour::insert("kernel32", "FindNextFileA", (void*&)oFindNextFileA, (void*)pFindNextFileA);
    detour::insert("kernel32", "GetFileSize", (void*&)oGetFileSize, (void*)pGetFileSize);
    detour::insert("kernel32", "SetFilePointer", (void*&)oSetFilePointer, (void*)pSetFilePointer);
    detour::insert("kernel32", "ReadFile", (void*&)oReadFile, (void*)pReadFile);
    detour::insert("kernel32", "WriteFile", (void*&)oWriteFile, (void*)pWriteFile);
  }

  if(dwReason == DLL_PROCESS_DETACH) {
    detour::remove("kernel32", "IsDebuggerPresent", (void*&)oIsDebuggerPresent);
    detour::remove("kernel32", "CreateFileA", (void*&)oCreateFileA);
    detour::remove("kernel32", "CloseHandle", (void*&)oCloseHandle);
    detour::remove("kernel32", "FindFirstFileA", (void*&)oFindFirstFileA);
    detour::remove("kernel32", "FindNextFileA", (void*&)oFindNextFileA);
    detour::remove("kernel32", "GetFileSize", (void*&)oGetFileSize);
    detour::remove("kernel32", "SetFilePointerA", (void*&)oSetFilePointer);
    detour::remove("kernel32", "ReadFile", (void*&)oReadFile);
    detour::remove("kernel32", "WriteFile", (void*&)oWriteFile);
  }

  return TRUE;
}
